/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Stack;
import NET.shudo.moba.reflect.*;


/**
 * Object serializer<BR>
 * java.io.ObjectOutputStream롣
 *
 * @author SHUDOH Kazuyuki
 */
public
class ObjectWriter extends DataOutputStream {
  /**
   * Ǥserializeobject
   */
  private Vector objectArray = new Vector();

  /**
   * Ǥserializeobjectɽ<BR>
   * object餽index롣<BR>
   * indexjava.lang.Integerinstance
   */
  private Dictionary objectDictionary = new Hashtable();

  private int nullCount = 0;


  public ObjectWriter(OutputStream out) {
    super(out);
  }

  public ObjectWriter() {
    super(System.out);
  }


  /**
   * ꥻåȤ롣˽񤭽Ф줿֥Ȥʤɤξ֤˺줵
   * ˺줿 ObjectWriter Ʊͤˤ롣
   */
  public void reset() {
    this.objectArray = new Vector();
    this.objectDictionary = new Hashtable();
    this.nullCount = 0;
  }


  /**
   * subclassΥ饹Ф򤹤뤿hook
   *
   * @return true if the specified class is the target class.
   */
  protected boolean isSpecificClass(Class clazz) {
    return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook<BR>
   * Υ饹Ȥ supeclass ˤĤƤǤ롣
   *
   * @return true if the specified class is the target class.
   */
  protected boolean writeSpecificClass(Class clazz, Object obj)
	throws IOException {
    return false;
  }


  public void writeObject(Object obj) throws IOException {
    try {
      writeObject0(obj);
      flushNull();
    }
    catch (StopMigrationException e) {
      // do nothing.
    }
  }

  private void writeObject0(Object obj) throws IOException {
//System.out.println("writeObject(" + obj + ") called.");
//System.out.flush();
    MetaClass metaClass = null;

    Stack metaClassStack = new Stack();
	/* instance var. ˤƤϤʤ */


    // null reference
    if (obj == null) {
      writeNull();
      return;
    }
    flushNull();

//System.out.println("class name: " + obj.getClass().getName());
//System.out.flush();
    // obj 
    if (obj instanceof String) {	// String#intern()
      obj = ((String)obj).intern();
    }


//    synchronized(objectArray) {
      // Ǥ serialize Ѥ
      Integer objNum = (Integer)objectDictionary.get(obj);
      if (objNum != null) {
//System.out.println("\tObject " + obj + " is already marshaled: " + objectArray.indexOf(obj));  System.out.flush();
	writeIndex(objNum.intValue());
	return;
      }

      // serialize ѤߤȤϿ  (줫 serialize Τ)
      objectArray.addElement(obj);
      objectDictionary.put(obj, new Integer(objectArray.size() - 1));
//System.out.println("\tObject " + obj + " is registered (index: " + (objectArray.size() - 1) + ").");
//    }


    metaClass = MetaClass.get(obj);


    // ̰륯饹
    if (checkSpecialClasses(obj))  return;


    if (metaClass.isArray()) {
      // 
      writeUTF(metaClass.signature());	// [L...;, [[..., [B, [I, ...

      if (metaClass.isObjectArray()) {	// ֥ or  
	writeObjectArray(obj);
      }
      else {				// ܷ
      // else if (metaClass.isBaseTypeArray()) {
	writeBaseTypeArray(obj);
      }

      return;
    }


    // 饹̾񤭽Ф
    writeUTF(obj.getClass().getName());

    // 饹 writeExternal() ư
    if (metaClass.isStreamable()) {
      ((Streamable)obj).writeToStream(this);
      return;
    }

    // MetaClass 򥹥åѤ
    while (metaClass != null) {
      metaClassStack.push(metaClass);
      if (isSpecificClass(metaClass.classClass()))
	break;
      metaClass = metaClass.superClass();
    }

    // java.lang.Object  subclass é
    while (!(metaClassStack.empty())) {
      metaClass = (MetaClass)metaClassStack.pop();

      // ObjectWriter  subclass ̰륯饹
      if (this.writeSpecificClass(metaClass.classClass(), obj))
	continue;

      // 饹 writeObject() ư
      if (metaClass.hasOwnWriteObject()) {
	invokeOwnObjectWriter(obj);

	write(0xff);	// terminater

	continue;
      }

      // ФϤ
      for (int i = 0; i < metaClass.fieldCount(); i++) {	// field é
	MetaField field = null;
	try{
	  field = metaClass.field(i);
	}
	catch (FieldNotFoundException e) {
	  // not reached
	  e.printStackTrace();  System.exit(1);
	}

	// serialize ʤե
	if (!field.isAccessible())  continue;

	// ֥ or 
	if (field.isObject()) {
	  writeObject(field.getObjectValue(obj));
	}
	// ܷ
	else {
	  writeBaseType(obj, field);
	}
      } // for (int i = 0; i < metaClass.fieldCount(); i++)  field é
    }	// while (!(metaClassStack.empty()))  subclass é


    flushNull();

    /* terminater */
    write(0xff);
    this.flush();
  }


  /**
   * 饹 writeObject() Ƥ֡
   */
  private native void invokeOwnObjectWriter(Object obj)  throws IOException;


  // for debug
  public void showObjectTable(PrintStream out) {
    int i = 0;
    for (Enumeration e = objectArray.elements(); e.hasMoreElements(); ) {
      out.println(i++ + ": " + (e.nextElement()));
    }
  }


  /**
   *  writeObject()
   */
  private void writeObject(ObjectWriter out) throws IOException {
  }
  /**
   *  readObject()
   */
  private void readObject(ObjectReader in)
	throws IOException, ClassNotFoundException {
  }


  /*
   * 
   */

  private final void writeNull()
	throws IOException {
    nullCount++;
  }

  private final void flushNull()
	throws IOException {
    if (nullCount <= 0)  return;

    if (nullCount <= 2) {
      for (int i = 0; i < nullCount; i++) {
	write(0xff);  write(0xff);
      }
    }
    else if (nullCount > 2) {
      write(0xff);  write(0xfe);
      writeShort(nullCount);
    }

    nullCount = 0;
  }


  private final void writeIndex(int index) throws IOException {
    flushNull();

    write(0xff);
    write(0xfd);
    writeInt(index);
  }

  /**
   * ̰륯饹
   */
  private boolean checkSpecialClasses(Object obj) throws IOException {
    // java.lang.String
    if (obj instanceof String) {
      writeString((String)obj);
      return true;
    }
    // java.lang.Class
    else if (obj instanceof Class) {
      writeClass((Class)obj);
      return true;
    }

    return false;
  }

  private final void writeString(String str) throws IOException {
    flushNull();

    write(0xff);
    write(0xfc);
    writeUTF(str);
  }

  private final void writeClass(Class clazz) throws IOException {
    flushNull();

    write(0xff);
    write(0xfb);
    writeUTF(clazz.getName());
  }


  /**
   * ܷ񤭽Ф
   */
  private final void writeBaseType(Object obj, MetaField field)
	throws IOException {
    String sig = field.signature();
    char c = sig.charAt(0);

    flushNull();

    switch(c) {
    case 'Z':
      writeBoolean(field.getBooleanValue(obj));
      break;
    case 'B':
      writeByte(field.getByteValue(obj));
      break;
    case 'C':
      writeChar(field.getCharValue(obj));
      break;
    case 'S':
      writeShort(field.getShortValue(obj));
      break;
    case 'I':
      writeInt(field.getIntValue(obj));
      break;
    case 'J':
      writeLong(field.getLongValue(obj));
      break;
    case 'F':
      writeFloat(field.getFloatValue(obj));
      break;
    case 'D':
      writeDouble(field.getDoubleValue(obj));
      break;
    default:
      // not reached
      // 顼  not implemented yet.
      break;
    }
  }

  /**
   * ֥Ȥ񤭽Ф
   */
  private final void writeObjectArray(Object obj)
	throws IOException {
    Object[] objArray = (Object[])obj;
    int length;

    if (objArray == null)  length = -1;
    else  length = objArray.length;

    flushNull();

    writeInt(length);	// Ĺ
    for (int j = 0; j < length; j++) {
      writeObject(objArray[j]);
    }
  }

  /**
   * ܷ񤭽Ф
   */
  private final native void writeBaseTypeArray(Object obj, String sig);

  /**
   * ܷ񤭽Ф
   */
  private final void writeBaseTypeArray(Object obj)
	throws IOException {
    MetaClass mc = MetaClass.get(obj);
    String sig = mc.signature();

    flushNull();

    writeBaseTypeArray(obj, sig);
  }
}
