/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for ClassWriter */

#include "NET_shudo_moba_io_ClassWriter.h"

#include <jni.h>	/* for native methods */

#include <native.h>	/* for native methods */

#include "moba.h"


/*
 * Local Functions
 */
static struct methodblock *metaMethodToMethodblock(
				JNIEnv *env, jobject /*MetaMethod*/ mm);


/* 饹طΥ᥽å */

/*
 * classfile ΥեޥåȤ interface Ϥ롣
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ClassWriter_writeIFInClassfile
  (JNIEnv *env, jobject _this, jclass clazz) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  int ifCount;
  short *interfaces;
  int i;

  ifCount = cbImplementsCount(cb);
  interfaces = cbImplements(cb);
#ifdef DEBUG
  printf("MetaClass#writeIFInClassfile() called.\n");
  fflush(stdout);
#endif

  for (i = 0; i < ifCount; i++) {
    if ((writeInt16(env, _this, (jshort *)(interfaces + i))) < 0)
      goto exc_occurred;
  }

  return;

exc_occurred:
  {
    jclass exc_clazz = (*env)->FindClass(env, "java/io/IOException");
    (*env)->ThrowNew(env, exc_clazz,
	"ClassWriter#writeIFInClassfile: write failed.");
  }
}


/*
 * ե֤̾
 */
JNIEXPORT jstring JNICALL Java_NET_shudo_moba_io_ClassWriter_sourceName
  (JNIEnv *env, jclass thisClazz, jclass clazz) {
  ClassClass *cb = (Hjava_lang_Class *)DeRef(env, clazz);
  char *sourceName;

  sourceName = cbSourceName(cb);

  if (!sourceName)  return NULL;

  return (*env)->NewStringUTF(env, sourceName);
}


/* ᥽åɴطΥ᥽å */

/*
 * max stack
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_maxStack
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->maxstack;
}

/*
 * max locals
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_maxLocals
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->nlocals;
}

/*
 * code length
 */
JNIEXPORT jlong JNICALL Java_NET_shudo_moba_io_ClassWriter_codeLength
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jlong)mb->code_length;
}

/*
 * ᥽å catch 㳰ο
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_catchFrameLength
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->exception_table_length;
}

/*
 * ᥽åǤ㳰 catch ˤĤƤξ
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_catchFrameStartPC
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->exception_table_length))  return -1;
  return (jint)(mb->exception_table)[index].start_pc;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_catchFrameEndPC
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->exception_table_length))  return -1;
  return (jint)(mb->exception_table)[index].end_pc;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_catchFrameHandlerPC
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->exception_table_length))  return -1;
  return (jint)(mb->exception_table)[index].handler_pc;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_catchFrameExceptionIndexInCP
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->exception_table_length))  return -1;
  return (jint)(mb->exception_table)[index].catchType;
}


/*
 * line no. table
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_lineNoTableLength
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->line_number_table_length;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_lineNoTablePC
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->line_number_table_length))  return -1;
  return (jint)(mb->line_number_table)[index].pc;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_lineNoTableLineNo
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->line_number_table_length))  return -1;
  return (jint)(mb->line_number_table)[index].line_number;
}


/*
 * local var. table
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTableLength
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->localvar_table_length;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTablePC
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->localvar_table_length))  return -1;
  return (jint)(mb->localvar_table)[index].pc0;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTableLife
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->localvar_table_length))  return -1;
  return (jint)(mb->localvar_table)[index].length;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTableNameIndex
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->localvar_table_length))  return -1;
  return (jint)(mb->localvar_table)[index].nameoff;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTableSigIndex
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->localvar_table_length))  return -1;
  return (jint)(mb->localvar_table)[index].sigoff;
}

JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_localVarTableSlot
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->localvar_table_length))  return -1;
  return (jint)(mb->localvar_table)[index].slot;
}


/*
 * ᥽åɤ throw 㳰ο
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_thrownExceptionLength
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm) {
#if (JDK_VER < 110)
  return 0;
#else
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  return (jint)mb->nexceptions;
#endif
}

/*
 * ᥽åɤ throw 㳰ˤĤƤξ
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_io_ClassWriter_thrownExceptionIndexInCP
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jint index) {
#if (JDK_VER < 110)
  return -1;
#else
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  if ((index < 0) || (index >= mb->nexceptions))  return -1;
  return (jint)(mb->exceptions)[index];
#endif
}


/*
 *  ((methodblock *)mb->code) Ϥ롣
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ClassWriter_writeCode
  (JNIEnv *env, jclass clazz, jobject /*MetaMethod*/ mm, jobject /*OutputStream*/ out) {
  struct methodblock *mb = metaMethodToMethodblock(env, mm);
  uint32_t length = (uint32_t)mb->code_length;
  unsigned char *code = mb->code;

  if ((length <= 0) || !code)  return;

  if ((writeBytes(env, out, (char *)code, (size_t)length)) < 0)
    goto exc_occurred;

  return;

exc_occurred:
  {
    jclass exc_clazz = (*env)->FindClass(env, "java/io/IOException");
    (*env)->ThrowNew(env, exc_clazz,
	"ClassWriter#writeCode: write failed.");
  }
}



/*
 *	Local Functions
 */

static struct methodblock *metaMethodToMethodblock(
				JNIEnv *env, jobject /*MetaMethod*/ obj) {
  jfieldID fid = NULL;
  jint methodID;

  if (!fid) {
    jclass clazz = (*env)->FindClass(env, "NET/shudo/moba/reflect/MetaMethod");
    fid = (*env)->GetFieldID(env, clazz, "methodID", "I");
  }

  methodID = (*env)->GetIntField(env, obj, fid);

  return (struct methodblock *)methodID;
}
