/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for marshaling */

#include <string.h>	/* for strlen() */
#ifdef __FreeBSD__
#  include <sys/types.h>	/* for type u_long in netinet/in.h */
#endif
#include <netinet/in.h>	/* for hton[ls],ntoh[ls]() */

#include "NET_shudo_moba_io_ObjectWriter.h"
#include "NET_shudo_moba_io_ObjectReader.h"

#include "moba.h"


/*
 * Local Functions
 */
static int writeArray(JNIEnv *env, jobject /*OutputStream*/ out,
			void *array, size_t size, size_t len);
static void fillArray(JNIEnv *env, jobject /*InputSteam*/ in,
			void *array, size_t size, size_t len);


/*
 *	Native methods for moba.ObjectOutputStream
 */

JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ObjectWriter_invokeOwnObjectWriter
  (JNIEnv *env, jobject _this, jobject obj) {
  jclass clazz = (*env)->GetObjectClass(env, obj);
  jmethodID mid = (*env)->GetMethodID(env, clazz,
		OWN_WRITEOBJ_METHOD, OWN_WRITEOBJ_SIGNATURE);

#if 0
  (*env)->CallVoidMethod(env, obj, mid, _this);
#else
  callMethod(env, obj, mid, 0, _this);
#endif
}


JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ObjectReader_invokeOwnObjectReader
  (JNIEnv *env, jobject _this, jobject obj) {
  jclass clazz = (*env)->GetObjectClass(env, obj);
  jmethodID mid = (*env)->GetMethodID(env, clazz,
		OWN_READOBJ_METHOD, OWN_READOBJ_SIGNATURE);

#if 0
  (*env)->CallVoidMethod(env, obj, mid, _this);
#else
  callMethod(env, obj, mid, 0, _this);
#endif
}


JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ObjectWriter_writeBaseTypeArray
  (JNIEnv *env, jobject _this, jobject obj, jstring sigStr) {
  jboolean sIsCopy;
  const char *sig;
  jint len;
  jboolean aIsCopy = JNI_FALSE;
  void *array;
  int ret = 0;

  sig = (*env)->GetStringUTFChars(env, sigStr, &sIsCopy);

  if (!obj) {
    len = -1;
    if ((writeInt32(env, _this, &len)) < 0) { ret = -1; goto finish_write; }
  }
  else {
    len = (*env)->GetArrayLength(env, (jarray)obj);
    if ((writeInt32(env, _this, &len)) < 0) { ret = -1; goto finish_write; }
  }

  if (sig[0] != '[') {	/* ǰΤ */
    ret = -1;
    goto finish_write;
  }

  switch (sig[1]) {
  case 'Z':
    array = (*env)->GetBooleanArrayElements(env, (jbooleanArray)obj,&aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jboolean), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseBooleanArrayElements(env,
		(jbooleanArray)obj, (jboolean *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'B':
    if ((writeArrayOfByte(env, _this, (jbyteArray)obj, 0, len)) < 0)
      ret = -1;
    goto finish_write;
    break;
  case 'C':
    array = (*env)->GetCharArrayElements(env, (jcharArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jchar), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseCharArrayElements(env,
		(jcharArray)obj, (jchar *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'S':
    array = (*env)->GetShortArrayElements(env, (jshortArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jshort), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseShortArrayElements(env,
		(jshortArray)obj, (jshort *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'I':
    array = (*env)->GetIntArrayElements(env, (jintArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jint), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseIntArrayElements(env,
		(jintArray)obj, (jint *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'J':
    array = (*env)->GetLongArrayElements(env, (jlongArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jlong), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseLongArrayElements(env,
		(jlongArray)obj, (jlong *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'F':
    array = (*env)->GetFloatArrayElements(env, (jfloatArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jfloat), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseFloatArrayElements(env,
		(jfloatArray)obj, (jfloat *)array, JNI_ABORT);
    goto finish_write;
    break;
  case 'D':
    array = (*env)->GetDoubleArrayElements(env, (jdoubleArray)obj, &aIsCopy);
    if ((writeArray(env, _this, array, sizeof(jdouble), (size_t)len)) < 0)
      ret = -1;
    if (aIsCopy)
      (*env)->ReleaseDoubleArrayElements(env,
		(jdoubleArray)obj, (jdouble *)array, JNI_ABORT);
    goto finish_write;
    break;
  default:
    /* not reached */
    ret = -1;
#ifdef DEBUG
    printf("writeBaseTypeArray(): invalid signature %s\n", sig);
    fflush(stdout);
#endif
    break;
  }

finish_write:
  /*  */
  if (sIsCopy)  (*env)->ReleaseStringUTFChars(env, sigStr, sig);

  if (ret < 0) {
    jclass exc_clazz = (*env)->FindClass(env, "java/io/IOException");
    (*env)->ThrowNew(env, exc_clazz,
	"ObjectWriter#writeBaseTypeArray: write failed.");
  }
}


JNIEXPORT void JNICALL Java_NET_shudo_moba_io_ObjectReader_fillBaseTypeArray
  (JNIEnv *env, jobject _this, jobject obj, jstring sigStr) {
  jboolean sIsCopy;
  const char *sig;
  jint len;
  jboolean aIsCopy = JNI_FALSE;
  void *array;

#ifdef DEBUG
  printf("ObjectReader#fillBaseTypeArray() is called.\n");  fflush(stdout);
#endif

  if (!obj) {
#ifdef DEBUG
    printf("ObjectReader#fillBaseTypeArray: object to be filled is null.\n");
    fflush(stdout);
#endif
    return;
  }

  len = (*env)->GetArrayLength(env, (jarray)obj);
#ifdef DEBUG
  printf("ObjectReader#fillBaseTypeArray: length: %d.\n", len);
  fflush(stdout);
#endif
  if (len < 0)  return;	/* ǰΤ */


  sig = (*env)->GetStringUTFChars(env, sigStr, &sIsCopy);
#ifdef DEBUG
  printf("ObjectReader#fillBaseTypeArray(): signature: %s\n", sig);
  fflush(stdout);
#endif
  if (sig[0] != '[')	/* ǰΤ */
    goto finish_read;
    

  switch (sig[1]) {
  case 'Z':
    array = (*env)->GetBooleanArrayElements(env, (jbooleanArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jboolean), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseBooleanArrayElements(env,
		(jbooleanArray)obj, (jboolean *)array, JNI_COMMIT);
    break;
  case 'B':
    readArrayOfByte(env, _this, (jbyteArray)obj, 0, len);
    break;
  case 'C':
    array = (*env)->GetCharArrayElements(env, (jcharArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jchar), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseCharArrayElements(env,
		(jcharArray)obj, (jchar *)array, JNI_COMMIT);
    break;
  case 'S':
    array = (*env)->GetShortArrayElements(env, (jshortArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jshort), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseShortArrayElements(env,
		(jshortArray)obj, (jshort *)array, JNI_COMMIT);
    break;
  case 'I':
    array = (*env)->GetIntArrayElements(env, (jintArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jint), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseIntArrayElements(env,
		(jintArray)obj, (jint *)array, JNI_COMMIT);
    break;
  case 'J':
    array = (*env)->GetLongArrayElements(env, (jlongArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jlong), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseLongArrayElements(env,
		(jlongArray)obj, (jlong *)array, JNI_COMMIT);
    break;
  case 'F':
    array = (*env)->GetFloatArrayElements(env, (jfloatArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jfloat), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseFloatArrayElements(env,
		(jfloatArray)obj, (jfloat *)array, JNI_COMMIT);
    break;
  case 'D':
    array = (*env)->GetDoubleArrayElements(env, (jdoubleArray)obj, &aIsCopy);
    fillArray(env, _this, array, sizeof(jdouble), (size_t)len);
    if (aIsCopy)
      (*env)->ReleaseDoubleArrayElements(env,
		(jdoubleArray)obj, (jdouble *)array, JNI_COMMIT);
    break;
  default:
    /* not reached */
#ifdef DEBUG
    printf("fillBaseTypeArray(): invalid signature %s\n", sig);
    fflush(stdout);
#endif
    break;
  }

finish_read:
  /*  */
  if (sIsCopy)  (*env)->ReleaseStringUTFChars(env, sigStr, sig);

#ifdef DEBUG
  printf("fillBaseTypeArray is done.\n");
  fflush(stdout);
#endif
  return;
}


/*
 * Local Functions
 */

static int writeArray(JNIEnv *env, jobject /*OutputStream*/ out,
			void *array, size_t size, size_t len) {
  int i;

  switch (size) {
  case 1:
    return writeBytes(env, out, (char *)array, len);
    break;
  case 2:
    {
      jshort *body = (jshort *)array;
      for (i = 0; i < len; i++) {
	if ((writeInt16(env, out, body)) < 0)  return -1;
	body++;
      }
    }
    break;
  case 4:
    {
      jint *body = (jint *)array;
      for (i = 0; i < len; i++) {
	if ((writeInt32(env, out, body)) < 0)  return -1;
	body++;
      }
    }
    break;
  case 8:
    {
      jlong *body = (jlong *)array;
printf("writeArray(): size = 8\n");  fflush(stdout);
      for (i = 0; i < len; i++) {
	if ((writeInt64(env, out, body)) < 0)  return -1;
	body++;
      }
    }
    break;
  default:
#ifdef DEBUG
    printf("writeArray(): invalid element size: %d\n", (int)size);
    fflush(stdout);
#endif
    break;
  }

  return 0;
}


static void fillArray(JNIEnv *env, jobject /*InputSteam*/ in,
			void *array, size_t size, size_t len) {
  int i;

  readBytes(env, in, (char *)array, size * len);

  switch (size) {
  case 1:
    break;
  case 2:
    {
      jshort *body = (jshort *)array;
      for (i = 0; i < len; i++) {
	*body = (jshort)ntohs(*(unsigned short *)body);
	body++;
      }
    }
    break;
  case 4:
    {
      jint *body = (jint *)array;
      for (i = 0; i < len; i++) {
	*body = (jint)ntohl(*(unsigned long *)body);
	body++;
      }
    }
    break;
  case 8:
    {
      jlong *body = (jlong *)array;
      for (i = 0; i < len; i++) {
	*body = (jlong)ntoh64(*(jlong *)body);
	body++;
      }
    }
    break;
  default:
#ifdef DEBUG
    printf("readArray(): invalid element size: %d\n", (int)size);
    fflush(stdout);
#endif
    break;
  }
}


