/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.reflect;

import java.lang.reflect.Field;


public
class MetaField {
  private Class thisClass;
  private Field thisField;
  private String thisSig;
  private String thisName;
  private int thisAccess;

  protected MetaField(Class clazz, Field field) {
    this.thisClass = clazz;
    this.thisField = field;

    this.thisSig = MetaClass.signatureByClass(field.getType());
    this.thisName = field.getName();
    this.thisAccess = field.getModifiers();
  }
  
  public String toString() {
    StringBuffer buf = new StringBuffer();

    buf.append(name());
    buf.append('{');
    buf.append(signature());
    buf.append('}');

    return buf.toString();
  }

  public String name() { return thisName; }
  public String signature() { return thisSig; }
  public int access() { return thisAccess; }

  /**
   * ΥեɤΥ饹롣
   *
   * @return Υեɤ֥ȤǤʤnull
   */
  public Class type() {  return thisField.getType();  }

  public boolean isAccessible() {
    int acc = this.access();

    if (((acc & Access.STATIC) | (acc & Access.TRANSIENT)) != 0)  return false;
    else  return true;
  };

  public boolean isArray() {  return Signature.isArray(signature());  }
  public boolean isObject() {  return Signature.isObject(signature());  }
  public boolean isObjectArray() {  return Signature.isObjectArray(signature());  }
  public boolean isBaseTypeArray() {  return Signature.isBaseTypeArray(signature());  }

  public boolean isAny() {
    return Signature.isAny(signature());
  }
  public boolean isByte() {
    return Signature.isByte(signature());
  }
  public boolean isChar() {
    return Signature.isChar(signature());
  }
  public boolean isDouble() {
    return Signature.isDouble(signature());
  }
  public boolean isEnum() {
    return Signature.isEnum(signature());
  }
  public boolean isFloat() {
    return Signature.isFloat(signature());
  }
  public boolean isInt() {
    return Signature.isInt(signature());
  }
  public boolean isLong() {
    return Signature.isLong(signature());
  }
  public boolean isShort() {
    return Signature.isShort(signature());
  }
  public boolean isVoid() {
    return Signature.isVoid(signature());
  }
  public boolean isBoolean() {
    return Signature.isBoolean(signature());
  }

  // եɤͤ
  public native byte getByteValue(Object obj);
  public native char getCharValue(Object obj);
  public native double getDoubleValue(Object obj);
  public native float getFloatValue(Object obj);
  public native int getIntValue(Object obj);
  public native long getLongValue(Object obj);
  public native short getShortValue(Object obj);
  public native boolean getBooleanValue(Object obj);
  public native Object getObjectValue(Object obj);

  public native byte getStaticByteValue();
  public native char getStaticCharValue();
  public native double getStaticDoubleValue();
  public native float getStaticFloatValue();
  public native int getStaticIntValue();
  public native long getStaticLongValue();
  public native short getStaticShortValue();
  public native boolean getStaticBooleanValue();
  public native Object getStaticObjectValue();

  // եɤͤ򥻥å
  public native void setByteValue(Object obj, byte val);
  public native void setCharValue(Object obj, char val);
  public native void setDoubleValue(Object obj, double val);
  public native void setFloatValue(Object obj, float val);
  public native void setIntValue(Object obj, int val);
  public native void setLongValue(Object obj, long val);
  public native void setShortValue(Object obj, short val);
  public native void setBooleanValue(Object obj, boolean val);
  public native void setObjectValue(Object obj, Object val);

  public native void setStaticByteValue(byte val);
  public native void setStaticCharValue(char val);
  public native void setStaticDoubleValue(double val);
  public native void setStaticFloatValue(float val);
  public native void setStaticIntValue(int val);
  public native void setStaticLongValue(long val);
  public native void setStaticShortValue(short val);
  public native void setStaticBooleanValue(boolean val);
  public native void setStaticObjectValue(Object val);
}
