/**
  Sample code for object serializer
  in package NET.shudo.moba.
  Copy this file to the directory which is in CLASSPATH and compile:
    % javac SerializeArrayTest.java
  and execute:
    % java SerializeArrayTest

  @author SHUDOH Kazuyuki
 */

import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;
import NET.shudo.moba.io.Streamable;


/**
 * class ObjectWriter, ObjectReader Υƥȡ
 */
public class SerializeArrayTest {
  private static final String OBJ_FILE = "output";
  private static final int ARRAY_SIZE = 10;

  private void run() {
    String[] strarray = new String[ARRAY_SIZE];
    strarray[3] = strarray[7] = "test string1";
    strarray[4] = strarray[8] = "test string2";

    double[] dblarray = new double[ARRAY_SIZE];
    dblarray[3] = dblarray[7] = 10.0;
    dblarray[4] = dblarray[8] = 100.0;

    int[] intarray = new int[ARRAY_SIZE];
    intarray[3] = intarray[7] = 5;
    intarray[4] = intarray[8] = 50;
    

    // output
    try {
      OutputStream fout = new FileOutputStream(OBJ_FILE);
      ObjectWriter sout = new ObjectWriter(fout);

      System.out.println("Writing to file \"" + OBJ_FILE + "\".");
      sout.writeObject(strarray);
      sout.writeObject(dblarray);
      sout.writeObject(strarray);
      sout.writeObject(intarray);

      sout.close();
    }
    catch (Exception e) {
      System.err.println(e); e.printStackTrace(); System.exit(1);
    }

    // input
    try {
      InputStream fin = new FileInputStream(OBJ_FILE);
      ObjectReader sin = new ObjectReader(fin);

      System.out.println("Reading from file \"" + OBJ_FILE + "\".");

      strarray = (String[])sin.readObject();
      dblarray = (double[])sin.readObject();

      for (int i = 0; i < ARRAY_SIZE; i++) {
	System.out.println("String[" + i + "]: " + strarray[i]);
      }
      for (int i = 0; i < ARRAY_SIZE; i++) {
	System.out.println("double[" + i + "]: " + dblarray[i]);
      }

      strarray = (String[])sin.readObject();

      for (int i = 0; i < ARRAY_SIZE; i++) {
	System.out.println("String[" + i + "]: " + strarray[i]);
      }

      intarray = (int[])sin.readObject();

      for (int i = 0; i < ARRAY_SIZE; i++) {
	System.out.println("int[" + i + "]: " + intarray[i]);
      }

      sin.close();
    }
    catch (Exception e) {
      System.err.println(e); e.printStackTrace(); System.exit(1);
    }

    System.out.println("Succeed!");
  }

  public static void main(String argv[]) {
    new SerializeArrayTest().run();
  }
}
